import textwrap

from langchain import LLMMathChain, OpenAI, SQLDatabase, SQLDatabaseChain
from langchain.agents import Tool, initialize_agent

from qabot.data_loader_chain import get_duckdb_data_loader_chain


def create_agent_executor(
        database_uri=None,
        database_engine=None,
        tables=None,
        return_intermediate_steps=False,
        callback_manager=None,
        verbose=False
):

    db_kwargs = dict(
        include_tables=tables,
        sample_rows_in_table_info=3
    )
    if database_engine is not None:
        db = SQLDatabase(database_engine, **db_kwargs)
    elif database_uri is not None:
        db = SQLDatabase.from_uri(
            database_uri,
            **db_kwargs
        )
    else:
        raise ValueError("Must provide either database_uri or database_engine")


    # Not quite working yet
    # llm = OpenAIChat(
    #     model_name="gpt-3.5-turbo",
    #     temperature=0.0
    # )

    llm = OpenAI(temperature=0.0)

    calculator_chain = LLMMathChain(llm=llm, verbose=False)
    db_chain = SQLDatabaseChain(llm=llm, database=db, verbose=True)

    data_loader_chain = get_duckdb_data_loader_chain(llm=llm, database=database_engine)

    tools = [
        #DuckDBTool(db),
        Tool(
            name="Calculator",
            func=calculator_chain.run,
            description="useful for when you need to answer questions about math"
        ),
        Tool(
            name="Data Lookup",
            func=db_chain.run,
            description=textwrap.dedent("""useful for when you need to answer questions requiring data. 
            Input should be in the form of a natural language question containing full context.
            
            """)
        ),
        Tool(
            name="Data Loader",
            func=data_loader_chain.run,
            description="useful for when you need to load data from a local or remote file. Files can be csv, json, sqlite, or parquet."
        )
    ]

    return initialize_agent(
        tools,
        llm,
        agent="zero-shot-react-description",
        callback_manager=callback_manager,
        return_intermediate_steps=return_intermediate_steps,
        #verbose=True
    )



from langchain import LLMChain, PromptTemplate
from langchain.agents import AgentExecutor, ZeroShotAgent

from qabot.data_loader_tool import DuckDBTool


def get_duckdb_data_query_chain(llm, database):
    tool = DuckDBTool(engine=database)

    prompt = PromptTemplate(
        input_variables=["input", "agent_scratchpad"],
        template=_DEFAULT_TEMPLATE,
    )

    llm_chain = LLMChain(llm=llm, prompt=prompt)
    tool_names = [tool.name]
    agent = ZeroShotAgent(llm_chain=llm_chain, allowed_tools=tool_names)
    agent_executor = AgentExecutor.from_agent_and_tools(
        agent=agent,
        tools=[tool],
        verbose=True
    )

    return agent_executor


_DEFAULT_TEMPLATE = """Given an input question, first create a syntactically correct DuckDB query to run,
then look at the results of the query and return the answer. 
 
Unless the user specifies in their question a specific number of examples to obtain, always limit your query
to at most {top_k} results. You can order the results by a relevant column to return the most interesting 
examples in the database.

Never query for all the columns from a specific table, only ask for a the few relevant columns given the question.

Pay attention to use only the column names that you can see in the schema description. Be careful to not query
for columns that do not exist. Also, pay attention to which column is in which table.

Use the following format:

Input: "Description of input data here"
Extension: Extension of the input data
Action: the action to take, should be one of [execute]
Action Input: "SQL Query to load data"
Final Answer: Name of table or view that was created

For example:
 
Input: "open test.parquet from the data folder"
Extension: parquet
Action: execute
Action Input: "CREATE TABLE test AS SELECT * FROM 'data/test.parquet';"
Final Answer: test

For example:
 
Input: "import data/records.json as customers"
Extension: json
Action: execute
Action Input: "CREATE table customers AS SELECT * FROM 'data/records.json';"
Final Answer: customers

Another Example:

Input: "create a view of the covid data from "s3://covid19-lake/enigma-jhu-timeseries/csv/jhu_csse_covid_19_timeseries_merged.csv"
Extension: csv
Action: execute
Action Input: "CREATE VIEW covid AS SELECT * FROM 's3://covid19-lake/enigma-jhu-timeseries/csv/jhu_csse_covid_19_timeseries_merged.csv';"
Final Answer: covid

{agent_scratchpad}

Input: {input}"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qabot']

package_data = \
{'': ['*']}

install_requires = \
['duckdb-engine>=0.6.9,<0.7.0',
 'langchain>=0.0.100,<0.0.101',
 'openai>=0.27.0,<0.28.0',
 'psycopg2-binary>=2.9.5,<3.0.0',
 'rich>=13.3.1,<14.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['qabot = qabot.cli:run']}

setup_kwargs = {
    'name': 'qabot',
    'version': '0.1.4',
    'description': '',
    'long_description': '# qabot\n\nQuery local or remote files or databases with natural language queries powered by\n`langchain` and `openai`.\n\nWorks on local CSV files:\n\n![](.github/local_csv_query.png)\n\nremote CSV files:\n\n```\n$ qabot \\\n    -f https://www.stats.govt.nz/assets/Uploads/Environmental-economic-accounts/Environmental-economic-accounts-data-to-2020/renewable-energy-stock-account-2007-2020-csv.csv \\\n    -q "How many Gigawatt hours of generation was there for Solar resources in 2015 through to 2020?"\n```\n\n\nas well as on real databases:\n\n![](.github/external_db_query.png)\n\nEven on (public) data stored in S3:\n\n![](.github/external_s3_data.png)\n\n\n## Quickstart\n\nYou need to set the `OPENAI_API_KEY` environment variable to your OpenAI API key, \nwhich you can get from [here](https://platform.openai.com/account/api-keys).\n\nInstall the `qabot` command line tool using pip/poetry:\n\n\n```bash\n$ poetry install qabot\n```\n\nThen run the `qabot` command with either files or a database connection string.\n\n## Examples\n\n### Local CSV file/s\n\n```bash\n$ qabot -q "how many passengers survived by gender?" -f data/titanic.csv\n🦆 Loading data from files...\nLoading data/titanic.csv into table titanic...\n\nQuery: how many passengers survived by gender?\nResult:\nThere were 233 female passengers and 109 male passengers who survived.\n\n\n 🚀 any further questions? [y/n] (y): y\n\n 🚀 Query: what was the largest family who did not survive? \nQuery: what was the largest family who did not survive?\nResult:\nThe largest family who did not survive was the Sage family, with 8 members.\n\n 🚀 any further questions? [y/n] (y): n\n```\n\n## Database\n\nInstall any required drivers for your database, e.g. `pip install psycopg2-binary` for postgres.\n\nFor example to connect and query directly from the trains database in the [relational dataset repository](https://relational.fit.cvut.cz/dataset/Trains):\n\n```bash\n$ pip install mysqlclient\n\n$ qabot -d mysql+mysqldb://guest:relational@relational.fit.cvut.cz:3306/trains -q "what are the unique load shapes of cars, what are the maximum number of cars per train?" \nQuery: what are the unique load shapes of cars, what are the maximum number of cars per train?\nResult:\nThe unique load shapes of cars are circle, diamond, hexagon, rectangle, and triangle, and the maximum number of cars per train is 3.\n```\n\nNote you can also supply a database connection string via the environment variable\n`QABOT_DATABASE_URI`.\n\n#### Limit the tables\n\nYou can limit the tables that are queried by passing the `-t` flag - this will save your tokens!\nFor example, to only query the `cars` table:\n\n```bash\n$ export QABOT_DATABASE_URI=mysql+mysqldb://guest:relational@relational.fit.cvut.cz:3306/trains\n$ qabot -q "what are the unique load shapes of cars?" -t cars\n\nThe unique load shapes of cars are circle, hexagon, triangle, rectangle, and diamond.\n```\n\n## Features\n\n## Intermediate steps and database queries\n\nUse the `-v` flag to see the intermediate steps and database queries:\n\n```bash\n$ qabot -d mysql+mysqldb://guest:relational@relational.fit.cvut.cz:3306/trains -q "what are the unique load shapes of cars, what are the maximum number of cars per train?" -v\nQuery: what are the unique load shapes of cars, what are the maximum number of cars per train?\nIntermediate Steps: \n  Step 1\n\n    list_tables_sql_db(\n      \n    )\n\n    Output:\n    trains, cars\n\n  Step 2\n\n    schema_sql_db(\n      trains, cars\n    )\n\n    Output:\n    CREATE TABLE trains (\n        id INTEGER(11) NOT NULL, \n        direction VARCHAR(4), \n        PRIMARY KEY (id)\n    )ENGINE=InnoDB DEFAULT CHARSET=latin1\n\n    SELECT * FROM \'trains\' LIMIT 3;\n    id  direction\n    1   east\n    2   east\n    3   east\n\n\n    CREATE TABLE cars (\n        id INTEGER(11) NOT NULL, \n        train_id INTEGER(11), \n        `position` INTEGER(11), \n        shape VARCHAR(255), \n        len VARCHAR(255), \n        sides VARCHAR(255), \n        roof VARCHAR(255), \n        wheels INTEGER(11), \n        load_shape VARCHAR(255), \n        load_num INTEGER(11), \n        PRIMARY KEY (id), \n        CONSTRAINT cars_ibfk_1 FOREIGN KEY(train_id) REFERENCES trains (id) ON DELETE CASCADE ON UPDATE CASCADE\n    )ENGINE=InnoDB DEFAULT CHARSET=latin1\n\n    SELECT * FROM \'cars\' LIMIT 3;\n    id  train_id        position        shape   len     sides   roof    wheels  load_shape      load_num\n    1   1       1       rectangle       short   not_double      none    2       circle  1\n    2   1       2       rectangle       long    not_double      none    3       hexagon 1\n    3   1       3       rectangle       short   not_double      peaked  2       triangle        1\n\n  Step 3\n\n    query_sql_db(\n      SELECT load_shape, MAX(load_num) FROM cars GROUP BY load_shape\n    )\n\n    Output:\n    [(\'circle\', 3), (\'diamond\', 1), (\'hexagon\', 1), (\'rectangle\', 3), (\'triangle\', 3)]\n\n\nResult:\nThe unique load shapes of cars are circle, diamond, hexagon, rectangle, and triangle, and the maximum number of cars per train is 3.\n\n```\n\n## Data accessed via http/s3\n\nUse the `-f <url>` flag to load data from a url, e.g. a csv file on s3:\n\n```bash\n$ qabot -f s3://covid19-lake/enigma-jhu-timeseries/csv/jhu_csse_covid_19_timeseries_merged.csv -q "how many confirmed cases are there" -v\n🦆 Loading data from files...\ncreate table jhu_csse_covid_19_timeseries_merged as select * from \'s3://covid19-lake/enigma-jhu-timeseries/csv/jhu_csse_covid_19_timeseries_merged.csv\';\n\nResult:\n264308334 confirmed cases\n```\n\n## Links\n\n- [Python library docs](https://langchain.readthedocs.io)\n- [Agent docs to talk to arbitrary apis via OpenAPI/Swagger](https://langchain.readthedocs.io/en/latest/modules/agents/agent_toolkits/openapi.html)\n- [Agents/Tools to talk SQL](https://langchain.readthedocs.io/en/latest/modules/agents/agent_toolkits/sql_database.html)\n- [Typescript library](https://hwchase17.github.io/langchainjs/docs/overview/)\n\n',
    'author': 'Brian Thorne',
    'author_email': 'brian@hardbyte.nz',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from typing import Any

from langchain.tools import BaseTool
from sqlalchemy import text


class DuckDBTool(BaseTool):
    name = "execute"
    description = """useful for when you need to run SQL queries against a DuckDB database."""

    database: Any = None

    def __init__(self, engine, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.database = engine

    def _run(self, query: str) -> str:
        try:
            res = self.database.execute(text(query))
            print(res)
            return res
        except Exception as e:
            return str(e)

    async def _arun(self, query: str) -> str:
        raise NotImplementedError("Data Loader does not support async")





import os
from urllib.parse import urlparse
from sqlalchemy import create_engine, text
from sqlalchemy.exc import ProgrammingError


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except:
        return False


def create_duckdb_from_files(files: list[str]):
    # By default, duckdb is fully in-memory - we can provide a path to get
    # persistent storage

    engine = create_engine("duckdb:///:memory:")

    for i, file_path in enumerate(files, 1):

        load_external_data_into_db(engine, file_path)

        # Alternative is to allow user to pass in column types:
        # conn.sql(f"""
        # create table {table_name} as (
        #     select * from read_csv_auto(
        #         '%s',
        #         delim='|',
        #         header=True,
        #         --types={'phone': 'VARCHAR'}
        #     )
        # )
        # """ % file_path)

    return engine


def load_external_data_into_db(engine, file_path, allow_view=False):
    # Work out if the filepath is actually a url (e.g. s3://)
    is_url = uri_validator(file_path)
    # Get the file name without extension from the file_path
    table_name, extension = os.path.splitext(os.path.basename(file_path))
    # If the table_name isn't a valid SQL identifier, we'll need to use something else
    try:
        engine.execute(text(f"create table t_{table_name} as select 1;"))
        engine.execute(text(f"drop table t_{table_name};"))
    except ProgrammingError as e:
        table_name = "data"

    # The SQLAgent doesn't appear to see view's just yet, so we'll create a table instead
    use_view = allow_view and is_url
    if is_url:
        engine.execute(text("INSTALL httpfs;"))
        engine.execute(text("LOAD httpfs;"))

    create_statement = f"create {'view' if use_view else 'table'} '{table_name}' as select * from '{file_path}';"
    engine.execute(text(create_statement))

